@@ -59,8 +59,11 @@
 #'
 fData = function( grid, values )
 {
-  all( abs( diff( unique( diff( grid ) ) ) ) < 1e-14 ) ||
-    stop( ' Error in fData: you provided an unevenly spaced grid')
+  if( length( unique( diff( grid ) ) ) > 1 )
+  {
+    max( diff( unique( diff( grid ) ) ) ) / diff( range( grid ) ) < 1e-13  ||
+      stop( ' Error in fData: you provided an unevenly spaced grid')
+  }
 
   P = length( grid )
 
@@ -700,6 +703,319 @@ mean.mfData = function( x, ... )
                   lapply( x$fDList, function( y )( mean( y )$values ) ) ) )
 }
 
+
+
+#' Covariance function for functional data
+#'
+#' \code{S3} method to compute the sample covariance and cross-covariance
+#' functions for a set of functional data.
+#'
+#' Given a univariate random function X, defined
+#' over the grid \eqn{I = [a,b]}, the covariance
+#' function is defined as:
+#'
+#' \deqn{C(s,t) = Cov( X(s), X(t) ), \qquad s,t \in I.}
+#'
+#' Given another random function, Y, defined over the same grid as X, the cross-
+#' covariance function of X and Y is:
+#'
+#' \deqn{C^{X,Y}( s,t ) =  Cov( X(s), Y(t) ), \qquad s, t \in I.}
+#'
+#' For a generic L-dimensional random function X, i.e. an L-dimensional
+#' multivariate functional datum, the covariance function is defined as the set
+#' of blocks:
+#'
+#' \deqn{C_{i,j}(s,t) = Cov( X_i(s), X_j(t)), i,j = 1, ...,L, s,t \in I,}
+#'
+#' while the cross-covariance function is defined by the blocks:
+#'
+#' \deqn{C^{X,Y}_{i,j}(s,t) = Cov( X_i(s), Y_j(t))}
+#'
+#' @details
+#'
+#' The method \code{cov_fun} provides the sample estimator of the covariance or
+#' cross-covariance functions for univariate or multivariate functional datasets.
+#'
+#' The class of \code{X} (\code{fData} or \code{mfData}) is used to dispatch the
+#' correct implementation of the method.
+#'
+#' @param X is the (eventually first) functional dataset, i.e. either an object
+#' of class \code{fData} or an object of class \code{mfData};
+#' @param Y is the (optional) second functional dataset to be used to compute the
+#' cross-covariance function,  either \code{NULL} or an \code{fData} or
+#' \code{mfData} object (see the Value section for details).
+#'
+#' @return
+#'
+#' The following cases are given:
+#'
+#' \itemize{
+#' \item{if \code{X} is of class \code{fData} and \code{Y} is \code{NULL}, then
+#' the covariance function of \code{X} is returned;}
+#' \item{if \code{X} is of class \code{fData} and \code{Y} is of
+#' class \code{fData},
+#' the cross-covariance function of the two datasets is returned;}
+#' \item{if \code{X} is of class \code{mfData} and \code{Y} is \code{NULL},
+#' the upper-triangular blocks of the covariance function of \code{X}
+#' are returned (in form of list and by row, i.e. in the squence 1_1, 1_2, ...,
+#' 1_L, 2_2, ... - have a look at the labels of the list with \code{str});}
+#' \item{if \code{X} is of class \code{mfData} and \code{Y} is of
+#' class \code{fData},
+#' the cross-covariances of \code{X}'s components and \code{Y} are
+#' returned (in form of list);}
+#' \item{if \code{X} is of class \code{mfData} and \code{Y} is of
+#' class \code{mfData},
+#' the upper-triangular blocks of the cross-covariance of \code{X}'s and
+#' \code{Y}'s components are returned (in form of list and by row, i.e. in the
+#' squence 1_1, 1_2, ..., 1_L, 2_2, ... - have a look at the labels
+#' of the list with \code{str}));}}
+#'
+#' In any case, the return type is either an instance of the \code{S3} class \code{Cov}
+#' or a list of instances of such class (for the case of multivariate
+#' functional data).
+#'
+#' @seealso \code{\link{fData}}, \code{\link{mfData}}, \code{\link{plot.Cov}}
+#'
+#' @examples
+#'
+#' # Generating a univariate functional dataset
+#' N = 1e2
+#'
+#' P = 1e2
+#' t0 = 0
+#' t1 = 1
+#'
+#' time_grid = seq( t0, t1, length.out = P )
+#'
+#' Cov = exp_cov_function( time_grid, alpha = 0.3, beta = 0.4 )
+#'
+#' D1 = generate_gauss_fdata( N, centerline = sin( 2 * pi * time_grid ), Cov = Cov )
+#' D2 = generate_gauss_fdata( N, centerline = sin( 2 * pi * time_grid ), Cov = Cov )
+#'
+#' fD1 = fData( time_grid, D1 )
+#' fD2 = fData( time_grid, D2 )
+#'
+#' # Computing the covariance function of fD1
+#'
+#' C = cov_fun( fD1 )
+#' str( C )
+#'
+#' # Computing the cross-covariance function of fD1 and fD2
+#' CC = cov_fun( fD1, fD2 )
+#' str( CC )
+#'
+#' # Generating a multivariate functional dataset
+#' L = 3
+#'
+#' C1 = exp_cov_function( time_grid, alpha = 0.1, beta = 0.2 )
+#' C2 = exp_cov_function( time_grid, alpha = 0.2, beta = 0.5 )
+#' C3 = exp_cov_function( time_grid, alpha = 0.3, beta = 1 )
+#'
+#' centerline = matrix( c( sin( 2 * pi * time_grid ),
+#'                         sqrt( time_grid ),
+#'                         10 * ( time_grid - 0.5 ) * time_grid ),
+#'                      nrow = 3, byrow = TRUE )
+#'
+#' D3 = generate_gauss_mfdata( N, L, centerline,
+#'                        correlations = c( 0.5, 0.5, 0.5 ),
+#'                        listCov = list( C1, C2, C3 ) )
+#'
+#' # adding names for better readability of BC3's labels
+#' names( D3 ) = c( 'comp1', 'comp2', 'comp3' )
+#' mfD3 = mfData( time_grid, D3 )
+#'
+#' D1 = generate_gauss_fdata( N, centerline = sin( 2 * pi * time_grid ),
+#'                               Cov = Cov )
+#' fD1 = fData( time_grid, D1 )
+#'
+#' # Computing the block covariance function of mfD3
+#' BC3 = cov_fun( mfD3 )
+#' str( BC3 )
+#'
+#' # computing cross-covariance between mfData and fData objects
+#' CC = cov_fun( mfD3, fD1 )
+#' str( CC )
+#'
+#' @export
+#'
+cov_fun = function( X, Y = NULL )
+{
+  UseMethod( 'cov_fun', X )
+}
+
+#' @rdname cov_fun
+#' @aliases cov_fun
+#'
+#' @importFrom stats cov
+#' @export
+cov_fun.fData = function( X, Y = NULL )
+{
+  if( ! is.null( Y ) )
+  {
+    if( class( Y ) != 'fData' )
+    {
+      stop( 'Error: you have to provide an fData object as Y argument')
+    }
+
+    if( X$t0 != Y$t0 || X$tP != Y$tP || X$h != Y$h ||
+        X$P != Y$P || X$N != Y$N )
+    {
+      stop( 'Error: you have to provide a pair of compliant fData objects')
+    }
+
+    return( structure( list( t0 = X$t0,
+                             tP = X$tP,
+                             h = X$h,
+                             P = X$P,
+                             values = stats::cov( X$values, Y$values ) ),
+                       class = 'Cov' ) )
+  } else {
+    return( structure( list( t0 = X$t0,
+                             tP = X$tP,
+                             h = X$h,
+                             P = X$P,
+                             values = stats::cov( X$values ) ),
+                       class = 'Cov' ) )
+  }
+}
+
+#' @rdname cov_fun
+#' @aliases cov_fun
+#'
+#' @export
+cov_fun.mfData = function( X, Y = NULL )
+{
+  if( ! is.null( Y ) )
+  {
+    if( ! class( Y ) %in% c( 'fData', 'mfData' ) )
+    {
+      stop( 'Error: you have to provide either an fData or mfData object')
+    } else if( X$N != Y$N || X$t0 != Y$t0 || X$tP != Y$tP || X$P != Y$P )
+    {
+      stop( 'Errror: you have to provide a Y dataset compliant to X')
+    }
+
+    if( class( Y ) == 'mfData' )
+    {
+      if( X$L != Y$L )
+        stop( 'You have to provide a Y dataset with same number of components as X')
+
+      list = NULL
+
+      for( i in 1 : X$L )
+      {
+        list = append( list,
+                       lapply( i : X$L,
+                               function( j ) cov_fun( X$fDList[[ i ]],
+                                                      Y$fDList[[ j ]] )  ) )
+      }
+
+      # Setting up names for the covariances
+      if( ! is.null( names( X$fDList ) ) )
+        nmx = names( X$fDList )
+      else nmx = 1 : X$L
+
+      if( ! is.null( names( Y$fDList ) ) )
+        nmy = names( Y$fDList )
+      else nmy = 1 : Y$L
+
+      nmlist = NULL
+
+      for( i in 1 : X$L )
+      {
+        nmlist = append( nmlist,
+                         sapply( i : X$L,
+                                 function( j ) ( paste( nmx[ i ],
+                                                        '_',
+                                                        nmy[ j ],
+                                                        sep = '' ) ) ) )
+      }
+      names( list ) = nmlist
+
+      return( list )
+
+    } else if( class( Y ) == 'fData' )
+    {
+      return( sapply( X = X$fDList, FUN = cov_fun, Y,
+                      simplify = FALSE, USE.NAMES = FALSE ) )
+    }
+  } else {
+
+    list = NULL
+
+    for( i in 1 : X$L )
+    {
+      list = append( list,
+                     lapply( i : X$L,
+                             function( j ) cov_fun( X$fDList[[ i ]],
+                                                    X$fDList[[ j ]] )  ) )
+    }
+
+    # Setting up names for the covariances
+    if( ! is.null( names( X$fDList ) ) )
+      nm = names( X$fDList )
+    else nm = 1 : X$L
+
+    nmlist = NULL
+
+    for( i in 1 : X$L )
+    {
+      nmlist = append( nmlist,
+                       sapply( i : X$L,
+                               function( j ) ( paste( nm[ i ],
+                                                      '_',
+                                                      nm[ j ],
+                                                      sep = '' ) ) ) )
+    }
+    names( list ) = nmlist
+
+    return( list )
+  }
+}
+
+#' Specialised method to plot \code{Cov} objects
+#'
+#' This function performs the plot of an object of class \code{Cov}, i.e. a
+#' covariance or cross-covaraince function.
+#'
+#' @details
+#' It builds above the function \code{graphics::image}, therefore any additional
+#' parameter suitable for \code{graphics::image} will also be suitable as \code{...}
+#' argument to \code{plot.Cov}.
+#'
+#' @param x the covariance or cross-covariance function of class \code{Cov}.
+#' @param ... additional graphical parameters to be used in plotting functions
+#'
+#' @seealso \code{\link{cov_fun}}
+#'
+#' @examples
+#'
+#' # Generating a univariate functional dataset
+#' N = 1e2
+#'
+#' P = 1e2
+#' t0 = 0
+#' t1 = 1
+#'
+#' time_grid = seq( t0, t1, length.out = P )
+#'
+#' Cov = exp_cov_function( time_grid, alpha = 0.3, beta = 0.4 )
+#'
+#' D1 = generate_gauss_fdata( N, centerline = sin( 2 * pi * time_grid ), Cov = Cov )
+#'
+#' fD1 = fData( time_grid, D1 )
+#'
+#' # Computing the covariance function of fD1
+#'
+#' plot( cov_fun( fD1 ), main = 'Covariance function', xlab = 'time', ylab = 'time' )
+#'
+#' @importFrom graphics image
+#' @export
+plot.Cov = function( x, ... )
+{
+  graphics::image( x$values, ...  )
+}
+
 #'
 #' Median of a univariate functional dataset
 #'
